// Copyright 2025 RnD Center "ELVEES", JSC

/*! \file
 *  \brief Заголовочный файл со структурами и функциями для управления вычислениями
 */

#ifndef _FFT
#define _FFT

#include "elcore50-signal-lib/DMAInit.h"
#include "elcore50-signal-lib/types_converter.h"

//! Тип данных преобразования
typedef enum {
    FLOAT = 1,      ///< float32
    FRACTIONAL = 2, ///< fractional16
    HFLOAT = 3      ///< float16
} DataTypes;

//! Максимальный размер сигнала, помещающегося во внутреннюю память
#define MAX_FFT_BUF_SIZE (512*2*16*4)
//! Размер буфера поворотов для максимального сигнала
#define MAX_FFT_W_BUF_SIZE ((3*512*4)/4)
//! Максимальный размер буфера во внутренней памяти
#define MAX_FFT_PLACE_SIZE (16384*2*4)

#pragma pack(push, 1)

//! Массивы данных, выгружаемых во внутреннюю память с dma
struct tdma_bufs {
    char x_d0[MAX_FFT_BUF_SIZE] __attribute__((aligned(32))); ///< первый входной блок
    char x_r0[MAX_FFT_BUF_SIZE] __attribute__((aligned(32))); ///< первый выходной блок
    char x_d1[MAX_FFT_BUF_SIZE] __attribute__((aligned(32))); ///< второй входной блок
    char x_r1[MAX_FFT_BUF_SIZE] __attribute__((aligned(32))); ///< второй выходной блок
    char t_d0[MAX_FFT_BUF_SIZE] __attribute__((aligned(32)));    ///< первый блок поворотов
    char t_d1[MAX_FFT_BUF_SIZE] __attribute__((aligned(32)));    ///< второй блок поворотов
    char w_d[MAX_FFT_W_BUF_SIZE] __attribute__((aligned(32)));   ///< блок поворотов столбцов/строк
};

//! Входной и выходной массивы во внутренней памяти
struct tinplace_bufs{
    char src[MAX_FFT_PLACE_SIZE] __attribute__((aligned(32))); ///< входной сигнал
    char dst[MAX_FFT_PLACE_SIZE] __attribute__((aligned(32))); ///< результат преобразования
};

#pragma pack(pop)

//! Структуры для вычислений во внутренней памяти
union tint_mem_bufs {
    struct tdma_bufs dma_bufs;          ///< для вычислений с dma
    struct tinplace_bufs inplace_bufs;  ///< для вычислений без dma
};

//! Размерность пространства
enum Dimension {
    D1 = 1,     ///< одномерное
    D2 = 2      ///< двумерное
};

//! Объект преобразования
typedef struct {
    int dim;            ///< размерность пространства
    int size;           ///< размер преобразования
    int height;         ///< количество строк матрицы при переходе к 2D представлению
    int length;         ///< количество столбцов матрицы при переходе к 2D представлению
    void *w_cols;       ///< вектор поворотов столбцов
    void *w_rows;       ///< вектор поворотов строк
    void *t;            ///< вектор поворотов промежуточного этапа
    void *src;          ///< входной сигнал
    void *dst;          ///< результат преобразования
    int dir;            ///< флаг типа преобразования: 1 - прямое, 0 - обратное
    int big_size;       ///< флаг преобразования большого размера: > 16K для типа float32, > 32K для типов fractional16 и float16
    DataTypes fft_type; ///< тип данных преобразования
    int element_size;   ///< размер типа в байтах
    int int_mem;        ///< флаг типа вычислений: 1 - во внутренней памяти, 0 - во внешней
    int ticks;      ///< количество тактов
    int instrs;      ///< количество инструкций
} fft_t;

//! Цепочки dma
typedef struct
{
    VDMAChain ch_in;        ///< цепочка входных блоков
    VDMAChain ch_out;       ///< цепочка выходных блоков
    VDMAChain ch_twiddles;  ///< цепочка блоков поворотов
} fft_dma_chains;

//! Тип вычислений
enum CALC_MEM {
    EXT_EXT = 0,    ///< во внешней памяти
    EXT_INT = 1     ///< во внутренней памяти
};
//! Тип преобразования
typedef enum {
    IFFT = 0,   ///< обратное
    FFT = 1     ///< прямое
} TransformType;

//! Оконные функции
typedef enum {
    HAMMING,            /// Непараметрическое окно Хемминга
    NUTTALL,            /// Непараметрическое окно Натталла
    BLACKMAN,           /// Непараметрическое окно Блэкмана
    BLACKMAN_HARRIS,    /// Непараметрическое окно Блэкмана-Харриса
    BLACKMAN_NUTTALL,   /// Непараметрическое окно Блэкмана-Натталла
    HANN,               /// Непараметрическое окно Ханна
    LANCZOS,            /// Непараметрическое окно Ланцоша
    BARTLETT,           /// Непараметрическое окно Бартлетта
    CHEBWIN             /// Параметрическое окно Дольф-Чебышева
} fir_window_type;
//! Типы фильтров
typedef enum {
    LOWPASS,            /// Фильтр нижних частот
    HIGHPASS,           /// Фильтр верхних частот
    BANDPASS,           /// Полосовой фильтр
    BANDSTOP            /// Режекторный фильтр
} filter_type;

// Error codes
#define SL_SUCCESS                            0
#define SL_MAX_SIZE_RESTRICTION_EXCEEDED     -1
#define SL_INVALID_POINTER                   -2
#define SL_INVALID_SIGNAL_SIZE               -3
#define SL_INVALID_DATA_TYPE                 -4
#define SL_INVALID_FILTER_TYPE               -5
#define SL_INVALID_WINDOW                    -6
#define SL_INVALID_CUTOFF_FREQUENCY          -7
#define SL_DMACHAIN_CREATION_FAILURE         -8

/// Обработка кодов возврата
void print_error_message(
    const int error_code     ///< [in] код возврата
);

/// Создание одномерного объекта преобразования
/// @return код статуса
int fft_plan_1d(
    fft_t* fft_obj,          ///< [in] входные данные типа fft_t
    int size,                ///< [in] количество комплексных элементов
    DataTypes fft_type       ///< [in] тип данных преобразования
);
/// Создание двумерного объекта преобразования
/// @return код статуса
int fft_plan_2d(
    fft_t *fft_obj,          ///< [in] входные данные типа fft_t
    const int height,        ///< [in] количество элементов в столбце
    const int length,        ///< [in] количество комплексных элементов в строке
    DataTypes fft_type       ///< [in] тип данных преобразовани
);
/// Управление вычислениями
/// @return код статуса
int fft_execute(
    fft_t* fft_obj           ///< [in] входные данные типа fft_t
);
/*!
 *  \fn void fft_free_sources(fft_t* fft_obj)
 *  \brief Очищение памяти после вычислений
 *  \param[in] fft_obj входные данные типа fft_t
 */
void fft_free_sources(fft_t* fft_obj);
/// Создание dma цепочки
/// @return код статуса
int create_vdma_fft_chain(
    const fft_t *fft_obj,     ///< [in] входные данные типа fft_t
    fft_dma_chains *chain     ///< [out] выходные данные типа fft_dma_chains
);
/// Вычисление одномерной быстрой БПФ-свертки
/// @return код статуса
int fft_conv_1d(
    float *src0,                     ///< [in] первый входной массив
    float *src1,                     ///< [in] второй входной массив
    float *dst,                      ///< [out] выходные данные
    const int32_t size,              ///< [in] количество комплексных элементов входного массива
    const int8_t is_data_in_xyram    ///< [in] где размещены входные данные
);
/// Вычисление двумерной быстрой БПФ-свертки
/// @return код статуса
int fft_conv_2d(
    float *src0,                      ///< [in] первый входной массив
    float *src1,                      ///< [in] второй входной массив
    float *dst,                       ///< [out] выходные данные
    const int32_t rows,               ///< [in] количество элементов в столбце
    const int32_t cols,               ///< [in] количество комплексных элементов в строке
    const int8_t is_data_in_xyram     ///< [in] где размещены входные данные
);

/// Вычисление функции оконного взвешивания
/// @return код статуса
int get_window(
    float *w,                            ///< [out] массив с вычисленной функцией
    const int32_t n,                     ///< [in] размер оконной функции (количество комплексных элементов)
    const fir_window_type win_type,      ///< [in] тип окна
    const int8_t is_symmetric,           ///< [in] симметричное или периодическое окно
    float param                          ///< [in] параметр для параметрического окна
);
/// Вычисление коэффициентов КИХ-фильтра нижних частот
/// @return код статуса
int fir_lpf(
    float *h,                            ///< [out] массив с вычисленными коэффициентами
    const int32_t ord,                   ///< [in] порядок фильтра
    const float wp,                      ///< [in] частота среза
    const fir_window_type win_type,      ///< [in] тип окна
    const int8_t is_symmetric,           ///< [in] симметричное или периодическое окно
    float param                          ///< [in] параметр для параметрического окна
);
/// Вычисление коэффициентов КИХ-фильтра методом оконного взвешивания
/// @return код статуса
int fir_window(
    float *h,                            ///< [out] массив с вычисленными коэффициентами
    const int32_t ord,                   ///< [in] порядок фильтра (количество комплексных элементов)
    const float w0,                      ///< [in] частота среза ФНЧ или левая частота среза для полосового и режекторного фильтров
    const float w1,                      ///< [in] частота среза ФВЧ или правая частота среза для полосового и режекторного фильтров
    const filter_type filter_type,       ///< [in] тип фильтра
    const fir_window_type win_type,      ///< [in] тип окна
    const int8_t is_symmetric,           ///< [in] симметричное или периодическое окно
    float param                          ///< [in] параметр для параметрического окна
);
/// Фильтрация сигнала с использованием цифровых фильтров
/// @return код статуса
int lfilter(
    float *b,                        ///< [in] коэффициенты числителя передаточной характеристики фильтра
    float *a,                        ///< [in] коэффициенты знаменателя передаточной характеристики фильтра
    const int16_t ord,               ///< [in] порядок фильтра
    float *x,                        ///< [in] входной сигнал для фильтрации
    const int16_t n,                 ///< [in] количество комплексных элементов во входном сигнале
    float *y,                        ///< [out] отфильтрованный сигнал
    const int8_t is_data_in_xyram    ///< [in] где размещены входные данные
);

/*!
 *  \fn void fft_float_cols_inter(void *dst, void *src, int height, const void *twiddles, int length, void *temp, int stage)
 *  \brief Вычисление БПФ по столбцам/строкам для типа float
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] height количество строк входной матрицы сигнала
 *  \param[in] twiddles вектор поворотов столбцов/строк
 *  \param[in] length количество столбцов входной матрицы сигнала
 *  \param[in] temp копия выходной матрицы
 *  \param[in] stage номер этапа преобразования
 */
extern void fft_float_cols_inter(void *dst, void *src, int height, const void *twiddles, int length,
                                 void *temp, int stage);
/*!
 *  \fn void fft_float_cols_inter_dma_tiles(void *dst, void *src, int height, void *twiddles, int length, void *temp, int stage)
 *  \brief Вычисление БПФ по столбцам/строкам для типа float с использованием dma
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] height количество строк входной матрицы сигнала
 *  \param[in] twiddles вектор поворотов столбцов/строк
 *  \param[in] length количество столбцов входной матрицы сигнала
 *  \param[in] temp копия выходной матрицы
 *  \param[in] stage номер этапа преобразования
 */
extern void fft_float_cols_inter_dma_tiles(void *dst, void *src, int height, void *twiddles,
                                           int length, void *temp, int stage);
/*!
 *  \fn void fft_fract_cols_inv(void *dst, void *src, int height, void *twiddles, int length, void *temp, int stage)
 *  \brief Вычисление обратного БПФ по столбцам/строкам для типа fractional16
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] height количество строк входной матрицы сигнала
 *  \param[in] twiddles вектор поворотов столбцов/строк
 *  \param[in] length количество столбцов входной матрицы сигнала
 *  \param[in] temp копия выходной матрицы
 *  \param[in] stage номер этапа преобразования
 */
extern void fft_fract_cols_inv(void *dst, void *src, int height, void *twiddles, int length,
                           void *temp, int stage);
/*!
 *  \fn void fft_fract_cols(void *dst, void *src, int height, void *twiddles, int length, void *temp, int stage)
 *  \brief Вычисление прямого БПФ по столбцам/строкам для типа fractional16
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] height количество строк входной матрицы сигнала
 *  \param[in] twiddles вектор поворотов столбцов/строк
 *  \param[in] length количество столбцов входной матрицы сигнала
 *  \param[in] temp копия выходной матрицы
 *  \param[in] stage номер этапа преобразования
 */
extern void fft_fract_cols(void *dst, void *src, int height, void *twiddles, int length,
                           void *temp, int stage);
/*!
 *  \fn void fft_hfloat_cols(void *dst, void *src, int height, void *twiddles, int length, void *temp, int stage)
 *  \brief Вычисление БПФ по столбцам/строкам для типа float16
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] height количество строк входной матрицы сигнала
 *  \param[in] twiddles вектор поворотов столбцов/строк
 *  \param[in] length количество столбцов входной матрицы сигнала
 *  \param[in] temp копия выходной матрицы
 *  \param[in] stage номер этапа преобразования
 */
extern void fft_hfloat_cols(void *dst, void *src, int height, void *twiddles, int length,
                            void *temp, int stage);

/*!
 *  \fn void fft_twiddles_cols_gen(void *w, int size)
 *  \brief Вычисление поворотов столбцов/строк для типа float
 *  \param[out] w выходные данные
 *  \param[in] size размер входных данных
 */
extern void fft_twiddles_cols_gen(void *w, int size);
/*!
 *  \fn void fft_twiddles_t_gen_inter(void *t, int size, int height, int length)
 *  \brief Вычисление поворотов для типа float
 *  \param[out] t выходные данные
 *  \param[in] size размер входных данных
 *  \param[in] height количество строк входной матрицы
 *  \param[in] length количество столбцов входной матрицы
 */
extern void fft_twiddles_t_gen_inter(void *t, int size, int height, int length);

/*!
 *  \fn void fft_fract_twiddles_cols_gen(void *w, int size)
 *  \brief Вычисление поворотов столбцов/строк для типа fractional16
 *  \param[out] w выходные данные
 *  \param[in] size размер входных данных
 */
extern void fft_fract_twiddles_cols_gen(void *w, int size);
/*!
 *  \fn void fft_fract_twiddles_t_gen(void *t, int size, int height, int length)
 *  \brief Вычисление поворотов для типа fractional16
 *  \param[out] t выходные данные
 *  \param[in] size размер входных данных
 *  \param[in] height количество строк входной матрицы
 *  \param[in] length количество столбцов входной матрицы
 */
extern void fft_fract_twiddles_t_gen(void *t, int size, int height, int length);

/*!
 *  \fn void fft_hfloat_twiddles_cols_gen(void  *w, int size)
 *  \brief Вычисление поворотов столбцов/строк для типа float16
 *  \param[out] w выходные данные
 *  \param[in] size размер входных данных
 */
extern void fft_hfloat_twiddles_cols_gen(void  *w, int size);
/*!
 *  \fn void fft_hfloat_twiddles_t_gen(void *t, int size, int height, int length)
 *  \brief Вычисление поворотов для типа float16
 *  \param[out] t выходные данные
 *  \param[in] size размер входных данных
 *  \param[in] height количество строк входной матрицы
 *  \param[in] length количество столбцов входной матрицы
 */
extern void fft_hfloat_twiddles_t_gen(void *t, int size, int height, int length);

/*!
 *  \fn void fft_mult_and_transpose_inter(void *dst, void *src, int src_height, int src_length, void *T)
 *  \brief Умножение на матрицу поворотов и транспонирование матрицы типа float
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 *  \param[in] T матрица поворотов
 */
extern void fft_mult_and_transpose_inter(void *dst, void *src, int src_height, int src_length, void *T);
/*!
 *  \fn void fft_mult_and_transpose(void  *dst, void  *src, int src_height, int src_length, void *T)
 *  \brief Умножение на матрицу поворотов и транспонирование матрицы типа float
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 *  \param[in] T матрица поворотов
 */
extern void fft_mult_and_transpose(void  *dst, void  *src, int src_height, int src_length, void *T);
/*!
 *  \fn void fft_mult_and_transpose_fract(void *dst, void  *src, int src_height, int src_length, void *T)
 *  \brief Умножение на матрицу поворотов и транспонирование матрицы типа fractional16
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 *  \param[in] T матрица поворотов
 */
extern void fft_mult_and_transpose_fract(void *dst, void  *src, int src_height, int src_length, void *T);
/*!
 *  \fn void fft_mult_and_transpose_hfloat(void  *dst, void  *src, int src_height, int src_length, void *T)
 *  \brief Умножение на матрицу поворотов и транспонирование матрицы типа float16
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 *  \param[in] T матрица поворотов
 */
extern void fft_mult_and_transpose_hfloat(void  *dst, void  *src, int src_height, int src_length, void *T);

/*!
 *  \fn void fft_transpose_inter(void *dst, void *src, int src_height, int src_length)
 *  \brief Транспонирование матрицы типа float
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 */
extern void fft_transpose_inter(void *dst, void *src, int src_height, int src_length);
/*!
 *  \fn void fft_transpose_fract(void *dst, void *src, int src_height, int src_length)
 *  \brief Транспонирование матрицы типа fractional16
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 */
extern void fft_transpose_fract(void *dst, void *src, int src_height, int src_length);
/*!
 *  \fn void fft_transpose_hfloat(void *dst, void *src, int src_height, int src_length)
 *  \brief Транспонирование матрицы типа float16
 *  \param[out] dst выходные данные
 *  \param[in] src входные данные
 *  \param[in] src_height количество строк входной матрицы
 *  \param[in] src_length количество столбцов входной матрицы
 */
extern void fft_transpose_hfloat(void *dst, void *src, int src_height, int src_length);

#endif
